// SPDX-FileCopyrightText: 2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "BookingEngine.h"

#include "SharedObjects/TournamentSettings.h"

#include "shared/Logging.h"
#include "shared/Json.h"

#include "network/ChecksumHelper.h"

// Qt includes
#include <QRandomGenerator>
#include <QJsonObject>
#include <QDebug>

// We use the whole ASCII set here, except for:
// space, single quotes, double quotes, backslash.
static const QString s_keySource = QLatin1String(
    "!#$%&()*+,-./"    // No space and no quoues
    "0123456789:;<=>?"
    "@ABCDEFGHIJKLMNO"
    "PQRSTUVWXYZ[]^_"  // No backslash
    "`abcdefghijklmno"
    "pqrstuvwxyz{|}~"
);

// CAUTION: If the settings version is bumped, the network protocol version also has to be,
// because the JSON serialization of the settings is transferred to a client as-is!
static constexpr int s_settingsVersion = 1;
static const QLatin1String s_settingsVersionKey("version");
static const QLatin1String s_tournamentName("tournament name");
static const QLatin1String s_securityKey("security key");

// QR data protocol

constexpr int s_protocolVersion = 1;

static const QString s_codeId = QLatin1String("MTBC"); // "MuckTurnier Booking Code"

static constexpr int s_checksumLength = 22; // MD5 hash, base64 encoded

const int BookingEngine::maxEscapedLength = 1295; // zz = 1295 in base 36

// Allow the whole printable ASCII range, except for:
//  - "%" (used for escaping)
//  - "`" and "^" (may be interpreted as a combining character)
//  - "\" (may be interpreted as an escaping character)
static const QByteArray s_escapeExcludes
    = QStringLiteral(" !\"#$&'()*+,-./:;<=>?@[]_{|}~").toUtf8();

// We swap non-ASCII characters likely to be found in names with ASCII-ones that can be used as-is
// to save space when the string is percent-encoded. After decoding, we swap the characters back.

static const QVector<QChar> s_possibleNameComponents = {
    QChar(228),  // "ä"
    QChar(246),  // "ö"
    QChar(252),  // "ü"
    QChar(196),  // "Ä"
    QChar(214),  // "Ö"
    QChar(220),  // "Ü"
    QChar(223),  // "ß"
    QChar(233),  // "é"
    QChar(232),  // "è"
    QChar(234),  // "ê"
    QChar(235),  // "ë"
    QChar(225),  // "á"
    QChar(224),  // "à"
    QChar(226),  // "â"
    QChar(243),  // "ó"
    QChar(242),  // "ò"
    QChar(241),  // "ñ"
    QChar(201),  // "É"
    QChar(200),  // "È"
    QChar(202),  // "Ê"
    QChar(193),  // "Á"
    QChar(8217), // "’" -- \u2019, a typographically correct apostrophe
};

static const QVector<QChar> s_replacementChars = {
    QChar::fromLatin1('!'),
    QChar::fromLatin1('"'),
    QChar::fromLatin1('#'),
    QChar::fromLatin1('$'),
    QChar::fromLatin1('('),
    QChar::fromLatin1(')'),
    QChar::fromLatin1('*'),
    QChar::fromLatin1(','),
    QChar::fromLatin1(':'),
    QChar::fromLatin1(';'),
    QChar::fromLatin1('<'),
    QChar::fromLatin1('='),
    QChar::fromLatin1('>'),
    QChar::fromLatin1('?'),
    QChar::fromLatin1('@'),
    QChar::fromLatin1('['),
    QChar::fromLatin1(']'),
    QChar::fromLatin1('_'),
    QChar::fromLatin1('{'),
    QChar::fromLatin1('|'),
    QChar::fromLatin1('}'),
    QChar::fromLatin1('~')
};

BookingEngine::BookingEngine(QObject *parent, TournamentSettings *tournamentSettings)
    : QObject(parent),
      m_tournamentSettings(tournamentSettings)
{
}

void BookingEngine::reload()
{
    m_tournamentName.clear();

    const auto settings = m_tournamentSettings->bookingSettings();
    const auto settingsEmpty = settings.isEmpty();
    m_tournamentSettings->setBookingEnabled(! settingsEmpty);

    if (settingsEmpty || ! parseSettings(settings)) {
        qCDebug(MuckturnierLog) << "BookingEngine: Initializing new security key";
        QString securityKey;
        const auto keySourceCount = s_keySource.count();
        for (int i = 0; i < 20; i++) {
            securityKey.append(s_keySource.at(
                QRandomGenerator::global()->bounded(keySourceCount)));
        }
        m_securityKey = securityKey;
    }
}

bool BookingEngine::parseSettings(const QString &settings)
{
    const auto object = Json::objectFromString(settings);

    const auto version = object.value(s_settingsVersionKey).toInt();
    if (version != s_settingsVersion) {
        qCWarning(MuckturnierLog) << "BookingEngine: Could not parse settings: Read invalid "
                                     "settings version: Got" << version << "expected"
                                  << s_settingsVersion;
        m_tournamentName.clear();
        m_securityKey.clear();
        return false;
    }

    const auto name = object.value(s_tournamentName).toString();
    if (name.isEmpty()) {
        qCWarning(MuckturnierLog) << "BookingEngine: Failed to parse settings: No tournament name "
                                     "set or invalid value for the tournament name";
        m_tournamentName.clear();
        m_securityKey.clear();
        return false;
    } else {
        m_tournamentName = name;
        m_escapedTournamentName = escape(name);
        if (m_escapedTournamentName.count() > BookingEngine::maxEscapedLength) {
            // This should not happen if the data wasn't messed up by hand
            qCWarning(MuckturnierLog) << "BookingEngine: Failed to parse settings: Tournament name "
                                         "is too long";
            m_tournamentName.clear();
            m_securityKey.clear();
            return false;
        }
    }

    const auto key = object.value(s_securityKey).toString();
    if (key.isEmpty()) {
        qCWarning(MuckturnierLog) << "BookingEngine: Failed to parse settings: No security key set "
                                     "or invalid value for the security key";
        m_tournamentName.clear();
        m_securityKey.clear();
        return false;
    } else {
        m_securityKey = key;
    }

    return true;
}

void BookingEngine::setTournamentName(const QString &name)
{
    m_tournamentName = name;
    m_escapedTournamentName = escape(name);
    Q_ASSERT_X(m_escapedTournamentName.count() <= BookingEngine::maxEscapedLength,
               "BookingEngine::setTournamentName", "The escaped tournament name is too long");
}

const QString &BookingEngine::tournamentName() const
{
    return m_tournamentName;
}

void BookingEngine::setSecurityKey(const QString &key)
{
    m_securityKey = key;
}

const QString &BookingEngine::securityKey() const
{
    return m_securityKey;
}

void BookingEngine::saveSettings()
{
    m_tournamentSettings->setBookingSettings(Json::serialize(QJsonObject {
        { s_settingsVersionKey, s_settingsVersion },
        { s_tournamentName,     m_tournamentName },
        { s_securityKey,        m_securityKey }
    }));

    m_tournamentSettings->setBookingEnabled(true);
}

QString BookingEngine::checksum(const QString &name, const QString &tournament) const
{
    ChecksumHelper checksumHelper;
    checksumHelper.addData(tournament.isEmpty() ? m_tournamentName : tournament);
    checksumHelper.addData(name);
    checksumHelper.addData(m_securityKey);
    return checksumHelper.checksum();
}

QString BookingEngine::swapReplacementChars(const QString &text) const
{
    auto swapped = text;
    const auto count = s_possibleNameComponents.count();
    for (int i = 0; i < swapped.count(); i++) {
        auto currentChar = swapped.at(i);
        for (int j = 0; j < count; j++) {
            if (currentChar == s_possibleNameComponents.at(j)) {
                swapped[i] = s_replacementChars.at(j);
            } else if (currentChar == s_replacementChars.at(j)) {
                swapped[i] = s_possibleNameComponents.at(j);
            }
        }
    }
    return swapped;
}

QString BookingEngine::escape(const QString &string) const
{
    // First swap possible name components with directly usable ASCII chars
    const auto swapped = swapReplacementChars(string).toUtf8();
    // Then percent-encode the rest
    return QString::fromUtf8(swapped.toPercentEncoding(s_escapeExcludes));
}

QString BookingEngine::unEscape(const QString &string) const
{
    // First percent-unencode the string
    const auto unescaped = QString::fromUtf8(QByteArray::fromPercentEncoding(string.toUtf8()));
    // Then un-swap the masked possible name components
    return swapReplacementChars(unescaped);
}

QByteArray BookingEngine::encodeInt(int value) const
{
    Q_ASSERT_X(value <= BookingEngine::maxEscapedLength,
               "BookingEngine::encodeInt", "Value bigger than maxEscapedLength passed");
    return QString::number(value, 36).rightJustified(2, QChar::fromLatin1('0')).toUtf8();
}

int BookingEngine::decodeInt(const QString &value) const
{
    Q_ASSERT_X(value.count() == 2, "BookingEngine::decodeInt", "Invalid value passed");
    bool okay;
    const auto result = value.toInt(&okay, 36);
    return okay ? result : -1;
}

QByteArray BookingEngine::encodeChecksum(const QString &checksum) const
{
    return QByteArray::fromHex(checksum.toUtf8()).toBase64(QByteArray::Base64Encoding
                                                           | QByteArray::OmitTrailingEquals);
}

QString BookingEngine::decodeChecksum(const QString &text) const
{
    const auto result = QByteArray::fromBase64Encoding(text.toUtf8());
    if (result.decodingStatus != QByteArray::Base64DecodingStatus::Ok) {
        return QString();
    }
    const auto checksum = QString::fromUtf8(result.decoded.toHex());
    if (checksum.count() != 32) { // The hex MD5 hash is 32 chars long
        return QString();
    }
    return checksum;
}

QByteArray BookingEngine::serialize(const QString &name, const QString &checksum) const
{
    QByteArray rawData;

    // Code ID header (4 chars)
    rawData.append(s_codeId.toUtf8());

    // Protocol version as hex (2 chars)
    rawData.append(encodeInt(s_protocolVersion));

    // Length of the escaped tournament name as hex (2 chars)
    rawData.append(encodeInt(m_escapedTournamentName.count()));

    // Length of the escaped pair/player name as hex (2 chars)
    const auto escapedName = escape(name);
    Q_ASSERT_X(escapedName.count() <= BookingEngine::maxEscapedLength,
               "BookingEngine::serialize", "Escaped name too long!");
    rawData.append(encodeInt(escapedName.count()));

    // The escaped tournament name
    rawData.append(m_escapedTournamentName.toUtf8());

    // The esacped pair/player name
    rawData.append(escapedName.toUtf8());

    // The checksum
    rawData.append(encodeChecksum(checksum));

    return rawData;
}

BookingEngine::QrData BookingEngine::parse(const QString &data) const
{
    const auto dataCount = data.count();

    QrData parsed;

    // Check the data to be valid

    // The smallest possible dataset is something like this:
    //
    // "MTBC 01 01 01 A B nNPitDtB4OEFJ8Sp9od5MA" (without the spaces)
    //
    // Header = 4 chars
    // Version = 2 chars
    // Length of tournament name data = 2 chars
    // Length of pair/player name data = 2 chars
    // Tournament name = at least 1 char
    // Pair/Player name = at least 1 char
    // Checksum = 22 chars
    //
    // In sum: 34 chars

    // Thus, the data is invalid if it's shorter than 34 chars:
    if (dataCount < 34) {
        parsed.error = tr("Der Datensatz ist zu kurz");
        return parsed;
    }

    // Check for our header (first four chars)
    if (data.left(4) != s_codeId) {
        parsed.error = tr("Der Datensatz enthält keinen gültigen Header");
        return parsed;
    }

    // Extract and check all numbers

    // Extract the version
    const auto version = decodeInt(data.mid(4, 2));
    if (version == -1) {
        parsed.error = tr("Die Protokoll-Versionsnummer konnte nicht ausgelesen werden");
        return parsed;
    }

    // Check for the version to be correct
    if (version != s_protocolVersion) {
        parsed.error = tr("Die Protokoll-Versionsnummer ist falsch "
                          "(erwartet: %1, gelesen: %2)").arg(
                          QString::number(s_protocolVersion), QString::number(version));
        return parsed;
    }

    // Extract the tournament name length
    const auto tournamentNameLength = decodeInt(data.mid(6, 2));
    if (tournamentNameLength == -1) {
        parsed.error = tr("Die Länge des Turniernamens konnte nicht ausgelesen werden");
        return parsed;
    }

    // Extract the players name length
    const auto nameLength = decodeInt(data.mid(8, 2));
    if (nameLength == -1) {
        parsed.error = tr("Die Länge des Paar- bzw. Spielernamens konnte nicht ausgelesen werden");
        return parsed;
    }

    // Do a plausibility check for the found lengths.
    // The header and the lengths are 10 chars long.
    // The checksum has s_checksumLength chars.
    // The rest has to be the tournament and the pair/player name.
    // Thus:
    if (dataCount - 10 - tournamentNameLength - nameLength != s_checksumLength) {
        parsed.error = tr("Die Längenangaben im Datensatz stimmen nicht mit der Datensatzlänge "
                          "überein");
        return parsed;
    }

    // Extract the tournament name
    const auto tournament = unEscape(data.mid(10, tournamentNameLength));

    // Extract the pair/player name
    const auto name = unEscape(data.mid(10 + tournamentNameLength, nameLength));

    // Extract the checksum
    const auto dataChecksum = decodeChecksum(data.right(s_checksumLength));
    if (dataChecksum.isEmpty()) {
        parsed.error = tr("Die Prüfsumme konnte nicht ausgelesen werden");
        return parsed;
    }

    // Fill in the data
    parsed.name = name;
    parsed.tournament = tournament;
    parsed.checksum = dataChecksum;
    parsed.isValid = dataChecksum == checksum(name, tournament);

    return parsed;
}

void BookingEngine::reset()
{
    m_tournamentSettings->setBookingSettings(QString());
    m_tournamentSettings->setBookingEnabled(false);
    reload();
}

bool BookingEngine::looksLikeBookingCode(const QString &string) const
{
    return string.startsWith(s_codeId);
}

bool BookingEngine::containsBookingCodeHeader(const QString &string) const
{
    return string.contains(s_codeId);
}
