// SPDX-FileCopyrightText: 2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "ImageLabel.h"

// Qt includes
#include <QDir>
#include <QResizeEvent>
#include <QTimer>

ImageLabel::ImageLabel(const QString &text, QWidget *parent)
    : QLabel(text, parent),
      m_initialText(text)
{
    setMinimumSize(128, 128);
    m_smoothTimer = new QTimer(this);
    m_smoothTimer->setSingleShot(true);
    connect(m_smoothTimer, &QTimer::timeout, this, &ImageLabel::setScaledImage);
}

void ImageLabel::removeImage()
{
    m_image = QImage();
    setScaledImage();
    setText(m_initialText);
    m_imagePath.clear();
}

void ImageLabel::setImage(const QString &file)
{
    if (file.isEmpty()) {
        removeImage();
        return;
    }

    m_image.load(file);
    if (m_image.isNull()) {
        setText(tr("Das Laden von\n%1\nist fehlgeschlagen!").arg(
                   QDir::toNativeSeparators(file)));
        setPixmap(QPixmap());
    } else {
        setText(QString());
        setScaledImage();
    }

    m_imagePath = file;
}

void ImageLabel::resizeEvent(QResizeEvent *event)
{
    if (m_image.isNull()) {
        return;
    }

    // During resizing, we use a quickly scaled image
    setPixmap(QPixmap::fromImage(m_image.scaled(size(), Qt::KeepAspectRatio)));

    // (Re)start the timer that sets a smoothly scaled version of the image
    m_smoothTimer->start(200);

    QLabel::resizeEvent(event);
}

void ImageLabel::setScaledImage()
{
    if (m_image.isNull()) {
        setPixmap(QPixmap());
        return;
    }

    setPixmap(QPixmap::fromImage(m_image.scaled(size(),
                                 Qt::KeepAspectRatio,
                                 Qt::SmoothTransformation)));
}

const QString &ImageLabel::imagePath() const
{
    return m_imagePath;
}
