// SPDX-FileCopyrightText: 2023-2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef SCHEDULEDISPLAY_H
#define SCHEDULEDISPLAY_H

// Qt includes
#include <QDialog>
#include <QHash>

// Local classes
class Settings;
class ResizableLabel;
class MarginSpinBox;
class ImageLabel;
class FontDialog;
class MuckturnierLabel;

// Qt classes
class QLabel;
class QJsonObject;
class QScreen;
class QStackedLayout;
class QMenu;

class ScheduleDisplay : public QDialog
{
    Q_OBJECT

public:
    enum Page {
        Welcome,
        Schedule,
        GoodBye
    };

    explicit ScheduleDisplay(QWidget *parent, Settings *settings, const QString &dbPath,
                             const QString &displayLogo);
    void showFullScreen(QScreen *screen);
    void setScheduleStarted(bool state);

public Q_SLOTS:
    void updateClock(const QString &text);
    void updateCurrentRound(int round);
    void updateNextRound(int round);
    void updateRoundTimeLeft(const QString &text);
    void updateTimeToNextRound(const QString &text);
    void updateNextRoundStart(const QString &text);
    void showPage(Page page);

private Q_SLOTS:
    void toggleFullScreen();
    void setFormattingVisible(bool state);
    void setMargin(Qt::Alignment pos, int margin);
    void chooseColor(const QString &name);
    void saveSettings();
    void resetSettings();
    void loadImage(Page page, const QString &file);
    void removeImage(Page page);
    void selectFont();

private: // Functions
    void loadSettings();
    void updateBackground();
    void setTimeToNextRoundText(const QString &text);

private: // Variables
    Settings *m_settings;
    const QString m_dbPath;

    FontDialog *m_fontDialog;

    QStackedLayout *m_layout;
    ImageLabel *m_welcome;
    QWidget *m_schedule;
    ImageLabel *m_goodBye;

    QMenu *m_menu;

    QWidget *m_labelsWrapper;
    QAction *m_setFormattingVisible;
    QAction *m_toggleFullScreen;

    QHash<QString, ResizableLabel *> m_labels;
    MuckturnierLabel *m_logo;
    QHash<Qt::Alignment, MarginSpinBox *> m_margins;
    QHash<Qt::Alignment, QLabel *> m_spacers;

    QHash<QString, QColor> m_colors;

    QFlags<Qt::WindowState> m_lastWindowState = Qt::WindowNoState;

    bool m_scheduleStarted = false;
    int m_currentRound = 0;
    int m_nextRound = 0;
    bool m_negativeTime = false;

};

#endif // SCHEDULEDISPLAY_H
