// SPDX-FileCopyrightText: 2010-2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "MainWindow.h"

#ifdef Q_OS_MACOS
#include "MacOSApplication.h"
#endif

#include "shared/Logging.h"

#include "SharedObjects/SharedObjects.h"
#include "SharedObjects/ResourceFinder.h"

// Qt includes

#include <QApplication>
#include <QDebug>
#include <QTranslator>
#include <QLibraryInfo>
#include <QFileInfo>
#include <QStyle>
#include <QPalette>

#ifndef Q_OS_MACOS
#include <QStringList>
#endif

int main(int argc, char *argv[])
{
    // Create the core application
#ifndef Q_OS_MACOS
    // On Linux and Windows, we can use a default QApplication:
    QApplication application(argc, argv);
#else
    // On macOS, "Open with" does not simply pass the file as a command line argument, but triggers
    // a QFileOpenEvent that we have to process. Thus, we need a custom QApplication on Mac:
    MacOSApplication application(argc, argv);
#endif

    // Detect if we use a dark theme
    const QPalette palette;
    const bool isDarkMode = palette.base().color().lightness()
                            < palette.windowText().color().lightness();
    application.setProperty("DARK_MODE", isDarkMode);

    // Set application-wide styles

    // Shared style
    QString style = QStringLiteral(
        "QGroupBox[checkable=false] { "
            "font-weight: bold; "
        "} "
    );

    const auto styleName = application.style()->objectName();

    // Fix the HLine and VLine QFrames appearance for Qt's standard styles
    // (Don't display those as a 1px black line, but use some gray color)
#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
    if (styleName == QLatin1String("fusion")
        || styleName.startsWith(QLatin1String("windows"))
        || styleName == QLatin1String("macintosh")) {

        // With Qt 5, we can use the "frameShape" selector, which currently won't work with Qt 6
        style.append(QStringLiteral(
            "QFrame[frameShape=\"4\"],"
            "QFrame[frameShape=\"5\"] { "
                "color: palette(Mid); "
            "} "
        ));
    }
#else
    if (styleName == QLatin1String("fusion")
        || styleName.startsWith(QLatin1String("windows"))
        || styleName == QLatin1String("macintosh")) {

        // This is a workaround for Bug #101877 (https://bugreports.qt.io/browse/QTBUG-101877)
        // which hopefully can be removed once the bug is actually fixed in Qt 6, along with all
        // setProperty("isseparator", true) calls on QFrames.
        style.append(QStringLiteral(
            "QFrame[isseparator=true] {"
                "border: 1px solid palette(Window); "
                "background-color: palette(Mid); "
            "} "
        ));
    }
#endif

    application.setStyleSheet(style);

    // Create the application-wide shared objects
    SharedObjects sharedObjects;
    auto *resourceFinder = sharedObjects.resourceFinder();

    // Install the translations

    // Nobody outside of Germany will use this, and nobody will ever translate
    // it to any other language. So let's just load the de locale stuff :-P

    // Add the Qt localizations

    QTranslator qtTranslator;

    // Look for a system-wide Qt locale file
    if (! qtTranslator.load(QStringLiteral("qtbase_de"),
                            QLibraryInfo::location(QLibraryInfo::TranslationsPath))) {

        qCDebug(MuckturnierLog) << "Could not load a system-wide Qt translation."
                                << "Trying to load a packaged one.";

        // Look for a resource file
        const auto path = resourceFinder->find(QStringLiteral("qtbase_de.qm"));
        if (path.isEmpty()) {
            qCDebug(MuckturnierLog) << "Could not find a packaged Qt translation!";
        } else if (! qtTranslator.load(QStringLiteral("qtbase_de"),
                                       QFileInfo(path).absolutePath())) {

            qCDebug(MuckturnierLog) << "Could not initialize the Qt translator!";
        }
    }

    if (! qtTranslator.isEmpty()) {
        application.installTranslator(&qtTranslator);
    }

    // Add the Muckturnier localizations (plural forms)

    QTranslator muckturnierTranslator;

    const QString path = resourceFinder->find(QStringLiteral("muckturnier_de.qm"));
    if (! path.isEmpty()
        && ! muckturnierTranslator.load(QStringLiteral("muckturnier_de"),
                                        QFileInfo(path).absolutePath())) {

        qCDebug(MuckturnierLog) << "Could not initialize the Muckturnier translator!";
    }

    if (! muckturnierTranslator.isEmpty()) {
        application.installTranslator(&muckturnierTranslator);
    }

    // Create the main window
    MainWindow mainWindow(&sharedObjects);
    mainWindow.show();

    // Handle "Open with ..." requests
#ifndef Q_OS_MACOS
    // On Linux and Windows, "Open with" simply passes the file to be opened as a command line
    // argument that we handle here:
    const QStringList arguments(application.arguments());
    if (arguments.count() > 1) {
        mainWindow.setEnabled(false);
        QApplication::processEvents();
        mainWindow.openTournament(arguments.at(1));
        mainWindow.setEnabled(true);
    }
#else
    // On macOS, we listen for a QFileOpenEvent in the custom QApplication, catch the file name to
    // be opened and pass it to the main window via the fileOpenRequested signal:
    QObject::connect(&application, &MacOSApplication::fileOpenRequested,
                     &mainWindow, &MainWindow::openTournament);
#endif

    return application.exec();
}
